@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2013-2021 Joris Vink <joris@coders.se>
+ * Copyright (c) 2013-2022 Joris Vink <joris@coders.se>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -30,10 +30,6 @@
 #include <netinet/in.h>
 #include <arpa/inet.h>
 
-#include <openssl/err.h>
-#include <openssl/dh.h>
-#include <openssl/ssl.h>
-
 #include <errno.h>
 #include <regex.h>
 #include <stdarg.h>
@@ -61,30 +57,33 @@ extern int daemon(int, int);
 #endif
 #endif
 
-/*
- * Figure out what type of OpenSSL API we are dealing with.
- */
-#if defined(LIBRESSL_VERSION_NUMBER)
-#if LIBRESSL_VERSION_NUMBER >= 0x3000000fL
-#define KORE_OPENSSL_NEWER_API		1
-#endif
-
-#if LIBRESSL_VERSION_NUMBER >= 0x3020200fL
-#define TLS1_3_VERSION			0x0304
-#endif
-
-#else
-#if OPENSSL_VERSION_NUMBER >= 0x10100000L
-#define KORE_OPENSSL_NEWER_API		1
-#endif
-#endif
-
 #if defined(__OpenBSD__)
 #define KORE_USE_PLATFORM_PLEDGE	1
 #endif
 
+#if defined(TLS_BACKEND_OPENSSL)
+#include <openssl/x509.h>
+#include <openssl/ssl.h>
+typedef X509		KORE_X509;
+typedef SSL		KORE_TLS;
+typedef SSL_CTX		KORE_TLS_CTX;
+typedef X509_NAME	KORE_X509_NAMES;
+typedef EVP_PKEY	KORE_PRIVATE_KEY;
+#else
+typedef void		KORE_X509;
+typedef void		KORE_TLS;
+typedef void		KORE_TLS_CTX;
+typedef void		KORE_X509_NAMES;
+typedef void		KORE_PRIVATE_KEY;
+#endif
+
 #define KORE_RSAKEY_BITS	4096
 
+/* Kore quit reasons. */
+#define KORE_QUIT_NONE		-1
+#define KORE_QUIT_NORMAL	0
+#define KORE_QUIT_FATAL		1
+
 #define KORE_RESULT_ERROR	0
 #define KORE_RESULT_OK		1
 #define KORE_RESULT_RETRY	2
@@ -100,18 +99,10 @@ extern int daemon(int, int);
 
 #define errno_s			strerror(errno)
 #define ssl_errno_s		ERR_error_string(ERR_get_error(), NULL)
-
 #define KORE_DOMAINNAME_LEN		255
 #define KORE_PIDFILE_DEFAULT		"kore.pid"
-#define KORE_DEFAULT_CIPHER_LIST	"ECDHE-RSA-AES128-GCM-SHA256:ECDHE-ECDSA-AES128-GCM-SHA256:ECDHE-RSA-AES256-GCM-SHA384:ECDHE-ECDSA-AES256-GCM-SHA384:kEDH+AESGCM:ECDHE-RSA-AES128-SHA256:ECDHE-ECDSA-AES128-SHA256:ECDHE-RSA-AES128-SHA:ECDHE-ECDSA-AES128-SHA:ECDHE-RSA-AES256-SHA384:ECDHE-ECDSA-AES256-SHA384:ECDHE-RSA-AES256-SHA:ECDHE-ECDSA-AES256-SHA:DHE-RSA-AES128-SHA256:DHE-RSA-AES128-SHA:DHE-RSA-AES256-SHA256:DHE-DSS-AES256-SHA:AES128-GCM-SHA256:AES256-GCM-SHA384:HIGH:!aNULL:!eNULL:!EXPORT:!DES:!3DES:!MD5:!PSK:!kRSA:!kDSA"
-
-#if defined(KORE_DEBUG)
-#define kore_debug(...)		\
-	if (kore_debug)		\
-		kore_debug_internal(__FILE__, __LINE__, __VA_ARGS__)
-#else
-#define kore_debug(...)
-#endif
+#define KORE_DHPARAM_PATH		PREFIX "/share/kore/ffdhe4096.pem"
+#define KORE_DEFAULT_CIPHER_LIST	"AEAD-AES256-GCM-SHA384:AEAD-CHACHA20-POLY1305-SHA256:AEAD-AES128-GCM-SHA256:ECDHE-RSA-AES256-GCM-SHA384:ECDHE-RSA-AES128-GCM-SHA256:DHE-RSA-AES256-GCM-SHA384:DHE-RSA-AES128-GCM-SHA256"
 
 #define NETBUF_RECV			0
 #define NETBUF_SEND			1
@@ -235,10 +226,11 @@ struct connection {
 	u_int8_t		state;
 	u_int8_t		proto;
 	struct listener		*owner;
-	X509			*cert;
-	SSL			*ssl;
+	KORE_TLS		*tls;
+	KORE_X509		*tls_cert;
 	char			*tls_sni;
 	int			tls_reneg;
+
 	u_int16_t		flags;
 	void			*hdlr_extra;
 
@@ -286,6 +278,9 @@ struct kore_runtime {
 	int	type;
 #if !defined(KORE_NO_HTTP)
 	int	(*http_request)(void *, struct http_request *);
+	void	(*http_request_free)(void *, struct http_request *);
+	void	(*http_body_chunk)(void *,
+		    struct http_request *, const void *, size_t);
 	int	(*validator)(void *, struct http_request *, const void *);
 	void	(*wsconnect)(void *, struct connection *);
 	void	(*wsdisconnect)(void *, struct connection *);
@@ -294,6 +289,7 @@ struct kore_runtime {
 #endif
 	void	(*execute)(void *);
 	int	(*onload)(void *, int);
+	void	(*signal)(void *, int);
 	void	(*connect)(void *, struct connection *);
 	void	(*configure)(void *, int, char **);
 };
@@ -303,6 +299,37 @@ struct kore_runtime_call {
 	struct kore_runtime	*runtime;
 };
 
+#if !defined(KORE_NO_HTTP)
+
+struct kore_route_params {
+	char			*name;
+	int			flags;
+	u_int8_t		method;
+	struct kore_validator	*validator;
+
+	TAILQ_ENTRY(kore_route_params)	list;
+};
+
+struct kore_route {
+	char					*path;
+	char					*func;
+	int					type;
+	int					errors;
+	int					methods;
+	regex_t					rctx;
+	struct kore_domain			*dom;
+	struct kore_auth			*auth;
+	struct kore_runtime_call		*rcall;
+	struct kore_runtime_call		*on_free;
+	struct kore_runtime_call		*on_headers;
+	struct kore_runtime_call		*on_body_chunk;
+
+	TAILQ_HEAD(, kore_route_params)		params;
+	TAILQ_ENTRY(kore_route)			list;
+};
+
+#endif
+
 struct kore_domain {
 	u_int16_t				id;
 	int					logerr;
@@ -323,10 +350,10 @@ struct kore_domain {
 	char					*crlfile;
 	char					*certfile;
 	char					*certkey;
-	SSL_CTX					*ssl_ctx;
+	KORE_TLS_CTX				*tls_ctx;
 	int					x509_verify_depth;
 #if !defined(KORE_NO_HTTP)
-	TAILQ_HEAD(, kore_module_handle)	handlers;
+	TAILQ_HEAD(, kore_route)		routes;
 	TAILQ_HEAD(, http_redirect)		redirects;
 #endif
 	TAILQ_ENTRY(kore_domain)		list;
@@ -362,15 +389,6 @@ LIST_HEAD(kore_server_list, kore_server);
 
 #define KORE_PARAMS_QUERY_STRING	0x0001
 
-struct kore_handler_params {
-	char			*name;
-	int			flags;
-	u_int8_t		method;
-	struct kore_validator	*validator;
-
-	TAILQ_ENTRY(kore_handler_params)	list;
-};
-
 #define KORE_AUTH_TYPE_COOKIE		1
 #define KORE_AUTH_TYPE_HEADER		2
 #define KORE_AUTH_TYPE_REQUEST		3
@@ -419,23 +437,6 @@ struct kore_module {
 	TAILQ_ENTRY(kore_module)	list;
 };
 
-#if !defined(KORE_NO_HTTP)
-
-struct kore_module_handle {
-	char					*path;
-	char					*func;
-	int					type;
-	int					errors;
-	regex_t					rctx;
-	struct kore_domain			*dom;
-	struct kore_runtime_call		*rcall;
-	struct kore_auth			*auth;
-	int					methods;
-	TAILQ_HEAD(, kore_handler_params)	params;
-	TAILQ_ENTRY(kore_module_handle)		list;
-};
-#endif
-
 /*
  * The workers get a 128KB log buffer per worker, and parent will fetch their
  * logs when it reached at least 75% of that or if its been > 1 second since
@@ -449,9 +450,17 @@ struct kore_alog_header {
 	u_int16_t		loglen;
 } __attribute__((packed));
 
+struct kore_privsep {
+	char		*root;
+	char		*runas;
+	int		skip_runas;
+	int		skip_chroot;
+};
+
 struct kore_worker {
 	u_int16_t			id;
 	u_int16_t			cpu;
+	int				ready;
 	int				running;
 #if defined(__linux__)
 	int				tracing;
@@ -462,7 +471,8 @@ struct kore_worker {
 	u_int8_t			has_lock;
 	int				restarted;
 	u_int64_t			time_locked;
-	struct kore_module_handle	*active_hdlr;
+	struct kore_route		*active_route;
+	struct kore_privsep		*ps;
 
 	/* Used by the workers to store accesslogs. */
 	struct {
@@ -556,13 +566,13 @@ struct kore_buf {
     kore_json_create_item(o, n, KORE_JSON_TYPE_STRING, v)
 
 #define kore_json_create_number(o, n, v)			\
-    kore_json_create_item(o, n, KORE_JSON_TYPE_NUMBER, v)
+    kore_json_create_item(o, n, KORE_JSON_TYPE_NUMBER, (double)v)
 
 #define kore_json_create_integer(o, n, v)			\
-    kore_json_create_item(o, n, KORE_JSON_TYPE_INTEGER, v)
+    kore_json_create_item(o, n, KORE_JSON_TYPE_INTEGER, (int64_t)v)
 
 #define kore_json_create_integer_u64(o, n, v)			\
-    kore_json_create_item(o, n, KORE_JSON_TYPE_INTEGER_U64, v)
+    kore_json_create_item(o, n, KORE_JSON_TYPE_INTEGER_U64, (u_int64_t)v)
 
 #define kore_json_create_literal(o, n, v)			\
     kore_json_create_item(o, n, KORE_JSON_TYPE_LITERAL, v)
@@ -570,7 +580,6 @@ struct kore_buf {
 struct kore_json {
 	const u_int8_t			*data;
 	int				depth;
-	int				error;
 	size_t				length;
 	size_t				offset;
 
@@ -588,12 +597,12 @@ struct kore_json_item {
 		char				*string;
 		double				number;
 		int				literal;
-		int64_t				s64;
+		int64_t				integer;
 		u_int64_t			u64;
 	} data;
 
-	int				(*parse)(struct kore_json *,
-					    struct kore_json_item *);
+	int	(*parse)(struct kore_json *,
+		    struct kore_json_item *);
 
 	TAILQ_ENTRY(kore_json_item)	list;
 };
@@ -659,6 +668,8 @@ struct kore_timer {
 #define KORE_MSG_CRL			9
 #define KORE_MSG_ACCEPT_AVAILABLE	10
 #define KORE_PYTHON_SEND_OBJ		11
+#define KORE_MSG_WORKER_LOG		12
+#define KORE_MSG_FATALX			13
 #define KORE_MSG_ACME_BASE		100
 
 /* messages for applications should start at 201. */
@@ -693,32 +704,28 @@ extern char	*config_file;
 #endif
 
 extern pid_t	kore_pid;
+extern int	kore_quit;
 extern int	kore_quiet;
-extern int	kore_debug;
 extern int	skip_chroot;
 extern int	skip_runas;
 extern int	kore_foreground;
 
 extern char	*kore_pidfile;
-extern char	*kore_root_path;
-extern char	*kore_runas_user;
-extern char	*kore_tls_cipher_list;
 
 extern volatile sig_atomic_t	sig_recv;
 
-extern int	tls_version;
-extern DH	*tls_dhparam;
-extern char	*rand_file;
-extern int	keymgr_active;
-extern char	*keymgr_runas_user;
-extern char	*keymgr_root_path;
-extern char	*acme_runas_user;
-extern char	*acme_root_path;
+extern char	*kore_rand_file;
+extern int	kore_keymgr_active;
+
+extern struct kore_privsep	worker_privsep;
+extern struct kore_privsep	keymgr_privsep;
+extern struct kore_privsep	acme_privsep;
 
 extern u_int8_t			nlisteners;
 extern u_int16_t		cpu_count;
 extern u_int8_t			worker_count;
 extern const char		*kore_version;
+extern const char		*kore_build_date;
 extern int			worker_policy;
 extern u_int8_t			worker_set_affinity;
 extern u_int32_t		worker_rlimit_nofiles;
@@ -734,26 +741,51 @@ extern struct kore_pool		nb_pool;
 extern struct kore_domain	*primary_dom;
 extern struct kore_server_list	kore_servers;
 
+/* kore.c */
 void		kore_signal(int);
 void		kore_shutdown(void);
+void		kore_signal_trap(int);
 void		kore_signal_setup(void);
 void		kore_proctitle(const char *);
 void		kore_default_getopt(int, char **);
 
+void		kore_server_closeall(void);
+void		kore_server_cleanup(void);
+void		kore_server_free(struct kore_server *);
+void		kore_server_finalize(struct kore_server *);
+
+struct kore_server	*kore_server_create(const char *);
+struct kore_server	*kore_server_lookup(const char *);
+
+void		kore_listener_accept(void *, int);
+struct listener	*kore_listener_lookup(const char *);
+void		kore_listener_free(struct listener *);
+struct listener	*kore_listener_create(struct kore_server *);
+int		kore_listener_init(struct listener *, int, const char *);
+
+int		kore_sockopt(int, int, int);
+int		kore_server_bind_unix(struct kore_server *,
+		    const char *, const char *);
+int		kore_server_bind(struct kore_server *,
+		    const char *, const char *, const char *);
+/* worker.c */
 void		kore_worker_reap(void);
-void		kore_worker_init(void);
+int		kore_worker_init(void);
+void		kore_worker_privsep(void);
+void		kore_worker_started(void);
 void		kore_worker_make_busy(void);
 void		kore_worker_shutdown(void);
 void		kore_worker_dispatch_signal(int);
-void		kore_worker_privdrop(const char *, const char *);
-void		kore_worker_spawn(u_int16_t, u_int16_t, u_int16_t);
+int		kore_worker_spawn(u_int16_t, u_int16_t, u_int16_t);
 int		kore_worker_keymgr_response_verify(struct kore_msg *,
 		    const void *, struct kore_domain **);
 
 void	kore_worker_entry(struct kore_worker *) __attribute__((noreturn));
 
 struct kore_worker	*kore_worker_data(u_int8_t);
+struct kore_worker	*kore_worker_data_byid(u_int16_t);
 
+/* platform code (linux.c, bsd.c) */
 void		kore_platform_init(void);
 void		kore_platform_sandbox(void);
 void		kore_platform_event_init(void);
@@ -781,12 +813,44 @@ void		kore_platform_pledge(void);
 void		kore_platform_add_pledge(const char *);
 #endif
 
+/* tls variants. */
+#define KORE_X509_NAME_COMMON_NAME	1
+
+void		kore_tls_init(void);
+void		kore_tls_cleanup(void);
+void		kore_tls_dh_check(void);
+int		kore_tls_supported(void);
+void		kore_tls_version_set(int);
+void		kore_tls_keymgr_init(void);
+int		kore_tls_dh_load(const char *);
+void		kore_tls_seed(const void *, size_t);
+int		kore_tls_ciphersuite_set(const char *);
+int		kore_tls_read(struct connection *, size_t *);
+void		kore_tls_domain_cleanup(struct kore_domain *);
+int		kore_tls_connection_accept(struct connection *);
+void		kore_tls_connection_cleanup(struct connection *);
+int		kore_tls_write(struct connection *, size_t, size_t *);
+void		kore_tls_domain_crl(struct kore_domain *, const void *, size_t);
+void		kore_tls_domain_setup(struct kore_domain *,
+		    int, const void *, size_t);
+
+KORE_PRIVATE_KEY	*kore_tls_rsakey_load(const char *);
+KORE_PRIVATE_KEY	*kore_tls_rsakey_generate(const char *);
+
+int		kore_tls_x509_data(struct connection *, u_int8_t **, size_t *);
+KORE_X509_NAMES	*kore_tls_x509_issuer_name(struct connection *);
+KORE_X509_NAMES	*kore_tls_x509_subject_name(struct connection *);
+int		kore_tls_x509name_foreach(KORE_X509_NAMES *, int, void *,
+		    int (*)(void *, int, int, const char *,
+		    const void *, size_t, int));
+/* accesslog.c */
 void		kore_accesslog_init(u_int16_t);
 void		kore_accesslog_worker_init(void);
 void		kore_accesslog_run(void *, u_int64_t);
 void		kore_accesslog_gather(void *, u_int64_t, int);
 
 #if !defined(KORE_NO_HTTP)
+/* auth.c */
 int		kore_auth_run(struct http_request *, struct kore_auth *);
 int		kore_auth_cookie(struct http_request *, struct kore_auth *);
 int		kore_auth_header(struct http_request *, struct kore_auth *);
@@ -796,6 +860,7 @@ int		kore_auth_new(const char *);
 struct kore_auth	*kore_auth_lookup(const char *);
 #endif
 
+/* timer.c */
 void		kore_timer_init(void);
 void		kore_timer_run(u_int64_t);
 u_int64_t	kore_timer_next_run(u_int64_t);
@@ -803,29 +868,7 @@ void		kore_timer_remove(struct kore_timer *);
 struct kore_timer	*kore_timer_add(void (*cb)(void *, u_int64_t),
 			    u_int64_t, void *, int);
 
-void		kore_server_closeall(void);
-void		kore_server_cleanup(void);
-void		kore_server_free(struct kore_server *);
-void		kore_server_finalize(struct kore_server *);
-
-struct kore_server	*kore_server_create(const char *);
-struct kore_server	*kore_server_lookup(const char *);
-
-void		kore_listener_accept(void *, int);
-struct listener	*kore_listener_lookup(const char *);
-void		kore_listener_free(struct listener *);
-struct listener	*kore_listener_create(struct kore_server *);
-int		kore_listener_init(struct listener *, int, const char *);
-
-int		kore_sockopt(int, int, int);
-int		kore_server_bind_unix(struct kore_server *,
-		    const char *, const char *);
-int		kore_server_bind(struct kore_server *,
-		    const char *, const char *, const char *);
-
-int		kore_tls_sni_cb(SSL *, int *, void *);
-void		kore_tls_info_callback(const SSL *, int, int);
-
+/* connection.c */
 void			kore_connection_init(void);
 void			kore_connection_cleanup(void);
 void			kore_connection_prune(int);
@@ -843,16 +886,19 @@ void			kore_connection_check_idletimer(u_int64_t,
 int			kore_connection_accept(struct listener *,
 			    struct connection **);
 
-u_int64_t	kore_time_ms(void);
 void		kore_log_init(void);
+void		kore_log_file(const char *);
 
 #if defined(KORE_USE_PYTHON)
 int		kore_configure_setting(const char *, char *);
 #endif
 
-void		*kore_malloc(size_t);
+/* config.c */
 void		kore_parse_config(void);
 void		kore_parse_config_file(FILE *);
+
+/* mem.c */
+void		*kore_malloc(size_t);
 void		*kore_calloc(size_t, size_t);
 void		*kore_realloc(void *, size_t);
 void		kore_free(void *);
@@ -860,15 +906,24 @@ void		kore_mem_init(void);
 void		kore_mem_cleanup(void);
 void		kore_mem_untag(void *);
 void		*kore_mem_lookup(u_int32_t);
+void		kore_mem_zero(void *, size_t);
 void		kore_mem_tag(void *, u_int32_t);
 void		*kore_malloc_tagged(size_t, u_int32_t);
 
+/* pool.c */
 void		*kore_pool_get(struct kore_pool *);
 void		kore_pool_put(struct kore_pool *, void *);
 void		kore_pool_init(struct kore_pool *, const char *,
 		    size_t, size_t);
 void		kore_pool_cleanup(struct kore_pool *);
 
+/* utils.c */
+void		fatal(const char *, ...) __attribute__((noreturn))
+		    __attribute__((format (printf, 1, 2)));
+void		fatalx(const char *, ...) __attribute__((noreturn))
+		    __attribute__((format (printf, 1, 2)));
+
+u_int64_t	kore_time_ms(void);
 char		*kore_time_to_date(time_t);
 char		*kore_strdup(const char *);
 time_t		kore_date_to_time(const char *);
@@ -879,22 +934,23 @@ size_t		kore_strlcpy(char *, const char *, const size_t);
 void		kore_server_disconnect(struct connection *);
 int		kore_split_string(char *, const char *, char **, size_t);
 void		kore_strip_chars(char *, const char, char **);
-int		kore_snprintf(char *, size_t, int *, const char *, ...);
+int		kore_snprintf(char *, size_t, int *, const char *, ...)
+		    __attribute__((format (printf, 4, 5)));
 long long	kore_strtonum(const char *, int, long long, long long, int *);
 double		kore_strtodouble(const char *, long double, long double, int *);
 int		kore_base64_encode(const void *, size_t, char **);
 int		kore_base64_decode(const char *, u_int8_t **, size_t *);
 int		kore_base64url_encode(const void *, size_t, char **, int);
 int		kore_base64url_decode(const char *, u_int8_t **, size_t *, int);
+int		kore_x509_issuer_name(struct connection *, char **, int);
+int		kore_x509_subject_name(struct connection *, char **, int);
+
 void		*kore_mem_find(void *, size_t, const void *, size_t);
 char		*kore_text_trim(char *, size_t);
 char		*kore_read_line(FILE *, char *, size_t);
 
-EVP_PKEY	*kore_rsakey_load(const char *);
-EVP_PKEY	*kore_rsakey_generate(const char *);
-int		kore_x509_subject_name(struct connection *, char **, int);
-
 #if !defined(KORE_NO_HTTP)
+/* websocket.c */
 void		kore_websocket_handshake(struct http_request *,
 		    const char *, const char *, const char *);
 int		kore_websocket_send_clean(struct netbuf *);
@@ -904,6 +960,7 @@ void		kore_websocket_broadcast(struct connection *,
 		    u_int8_t, const void *, size_t, int);
 #endif
 
+/* msg.c */
 void		kore_msg_init(void);
 void		kore_msg_worker_init(void);
 void		kore_msg_parent_init(void);
@@ -915,21 +972,27 @@ int		kore_msg_register(u_int8_t,
 		    void (*cb)(struct kore_msg *, const void *));
 
 #if !defined(KORE_NO_HTTP)
+/* filemap.c */
 void		kore_filemap_init(void);
 void		kore_filemap_resolve_paths(void);
-int		kore_filemap_create(struct kore_domain *, const char *,
-		    const char *);
 extern char	*kore_filemap_ext;
 extern char	*kore_filemap_index;
+
+struct kore_route	*kore_filemap_create(struct kore_domain *, const char *,
+			    const char *, const char *);
 #endif
 
+/* fileref.c */
 void			kore_fileref_init(void);
 struct kore_fileref	*kore_fileref_get(const char *, int);
 struct kore_fileref	*kore_fileref_create(struct kore_server *,
 			    const char *, int, off_t, struct timespec *);
 void			kore_fileref_release(struct kore_fileref *);
 
+/* domain.c */
 struct kore_domain	*kore_domain_new(const char *);
+struct kore_domain	*kore_domain_byid(u_int16_t);
+struct kore_domain	*kore_domain_lookup(struct kore_server *, const char *);
 
 void		kore_domain_init(void);
 void		kore_domain_cleanup(void);
@@ -945,28 +1008,36 @@ void		kore_domain_load_crl(void);
 void		kore_domain_keymgr_init(void);
 void		kore_domain_callback(void (*cb)(struct kore_domain *));
 int		kore_domain_attach(struct kore_domain *, struct kore_server *);
-void		kore_domain_tlsinit(struct kore_domain *, int,
-		    const void *, size_t);
-void		kore_domain_crl_add(struct kore_domain *, const void *, size_t);
+
 #if !defined(KORE_NO_HTTP)
-int		kore_module_handler_new(struct kore_domain *, const char *,
-		    const char *, const char *, int);
-void		kore_module_handler_free(struct kore_module_handle *);
-struct kore_module_handle	*kore_module_handler_find(struct http_request *,
-				    struct kore_domain *);
+/* route.c */
+void		kore_route_reload(void);
+void		kore_route_free(struct kore_route *);
+void		kore_route_callback(struct kore_route *, const char *);
+
+struct kore_route	*kore_route_create(struct kore_domain *,
+			    const char *, int);
+int			kore_route_lookup(struct http_request *,
+			    struct kore_domain *, int, struct kore_route **);
 #endif
 
+/* runtime.c */
 struct kore_runtime_call	*kore_runtime_getcall(const char *);
 struct kore_module		*kore_module_load(const char *,
 				    const char *, int);
 
 void	kore_runtime_execute(struct kore_runtime_call *);
 int	kore_runtime_onload(struct kore_runtime_call *, int);
+void	kore_runtime_signal(struct kore_runtime_call *, int);
 void	kore_runtime_configure(struct kore_runtime_call *, int, char **);
 void	kore_runtime_connect(struct kore_runtime_call *, struct connection *);
 #if !defined(KORE_NO_HTTP)
 int	kore_runtime_http_request(struct kore_runtime_call *,
 	    struct http_request *);
+void	kore_runtime_http_request_free(struct kore_runtime_call *,
+	    struct http_request *);
+void	kore_runtime_http_body_chunk(struct kore_runtime_call *,
+	    struct http_request *, const void *, size_t);
 int	kore_runtime_validator(struct kore_runtime_call *,
 	    struct http_request *, const void *);
 void	kore_runtime_wsconnect(struct kore_runtime_call *, struct connection *);
@@ -976,10 +1047,8 @@ void	kore_runtime_wsmessage(struct kore_runtime_call *,
 	    struct connection *, u_int8_t, const void *, size_t);
 #endif
 
-struct kore_domain	*kore_domain_byid(u_int16_t);
-struct kore_domain	*kore_domain_lookup(struct kore_server *, const char *);
-
 #if !defined(KORE_NO_HTTP)
+/* validator.c */
 void		kore_validator_init(void);
 void		kore_validator_reload(void);
 int		kore_validator_add(const char *, u_int8_t, const char *);
@@ -989,12 +1058,9 @@ int		kore_validator_check(struct http_request *,
 struct kore_validator	*kore_validator_lookup(const char *);
 #endif
 
-void		fatal(const char *, ...) __attribute__((noreturn));
-void		fatalx(const char *, ...) __attribute__((noreturn));
-
 const char	*kore_worker_name(int);
-void		kore_debug_internal(char *, int, const char *, ...);
 
+/* net.c */
 u_int16_t	net_read16(u_int8_t *);
 u_int32_t	net_read32(u_int8_t *);
 u_int64_t	net_read64(u_int8_t *);
@@ -1009,9 +1075,7 @@ int		net_send(struct connection *);
 int		net_send_flush(struct connection *);
 int		net_recv_flush(struct connection *);
 int		net_read(struct connection *, size_t *);
-int		net_read_tls(struct connection *, size_t *);
 int		net_write(struct connection *, size_t, size_t *);
-int		net_write_tls(struct connection *, size_t, size_t *);
 void		net_recv_reset(struct connection *, size_t,
 		    int (*cb)(struct netbuf *));
 void		net_remove_netbuf(struct connection *, struct netbuf *);
@@ -1024,6 +1088,7 @@ void		net_send_stream(struct connection *, void *,
 		    size_t, int (*cb)(struct netbuf *), struct netbuf **);
 void		net_send_fileref(struct connection *, struct kore_fileref *);
 
+/* buf.c */
 void		kore_buf_free(struct kore_buf *);
 struct kore_buf	*kore_buf_alloc(size_t);
 void		kore_buf_init(struct kore_buf *, size_t);
@@ -1033,23 +1098,29 @@ void		kore_buf_reset(struct kore_buf *);
 void		kore_buf_cleanup(struct kore_buf *);
 
 char	*kore_buf_stringify(struct kore_buf *, size_t *);
-void	kore_buf_appendf(struct kore_buf *, const char *, ...);
-void	kore_buf_appendv(struct kore_buf *, const char *, va_list);
+void	kore_buf_appendf(struct kore_buf *, const char *, ...)
+		    __attribute__((format (printf, 2, 3)));
+void	kore_buf_appendv(struct kore_buf *, const char *, va_list)
+		    __attribute__((format (printf, 2, 0)));
 void	kore_buf_replace_string(struct kore_buf *,
 	    const char *, const void *, size_t);
 
+/* json.c */
+int	kore_json_errno(void);
 int	kore_json_parse(struct kore_json *);
 void	kore_json_cleanup(struct kore_json *);
 void	kore_json_item_free(struct kore_json_item *);
 void	kore_json_init(struct kore_json *, const void *, size_t);
 void	kore_json_item_tobuf(struct kore_json_item *, struct kore_buf *);
+void	kore_json_item_attach(struct kore_json_item *, struct kore_json_item *);
 
-const char		*kore_json_strerror(struct kore_json *);
+const char		*kore_json_strerror(void);
 struct kore_json_item	*kore_json_find(struct kore_json_item *,
 			    const char *, u_int32_t);
 struct kore_json_item	*kore_json_create_item(struct kore_json_item *,
 			    const char *, u_int32_t, ...);
 
+/* keymgr.c */
 void	kore_keymgr_run(void);
 void	kore_keymgr_cleanup(int);
 
