@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2019-2021 Joris Vink <joris@coders.se>
+ * Copyright (c) 2019-2022 Joris Vink <joris@coders.se>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -262,13 +262,10 @@ static char		*revoke_url = NULL;
 static char		*account_id = NULL;
 static char		*account_url = NULL;
 
-static u_int8_t acme_alpn_name[] =
-    { 0xa, 'a', 'c', 'm', 'e', '-', 't', 'l', 's', '/', '1' };
-
+struct kore_privsep	acme_privsep;
+int			acme_domains = 0;
 char			*acme_email = NULL;
 char			*acme_provider = NULL;
-char			*acme_root_path = NULL;
-char			*acme_runas_user = NULL;
 u_int32_t		acme_request_timeout = 8;
 
 void
@@ -310,7 +307,7 @@ kore_acme_run(void)
 #if defined(KORE_USE_PYTHON)
 	kore_msg_unregister(KORE_PYTHON_SEND_OBJ);
 #endif
-	kore_worker_privdrop(acme_runas_user, acme_root_path);
+	kore_worker_privsep();
 
 #if defined(__OpenBSD__)
 	if (unveil("/etc/ssl/", "r") == -1)
@@ -321,14 +318,10 @@ kore_acme_run(void)
 
 	http_init();
 
-	if (!kore_quiet) {
-		kore_log(LOG_NOTICE,
-		    "acme worker started (pid#%d)", worker->pid);
-	}
-
 	LIST_INIT(&orders);
 	LIST_INIT(&signops);
 
+	kore_worker_started();
 	acme_parse_directory();
 
 	while (quit != 1) {
@@ -362,39 +355,6 @@ kore_acme_run(void)
 	net_cleanup();
 }
 
-int
-kore_acme_tls_alpn(SSL *ssl, const unsigned char **out, unsigned char *outlen,
-    const unsigned char *in, unsigned int inlen, void *udata)
-{
-	struct connection	*c;
-
-	if ((c = SSL_get_ex_data(ssl, 0)) == NULL)
-		fatal("%s: no connection data present", __func__);
-
-	if (inlen != sizeof(acme_alpn_name))
-		return (SSL_TLSEXT_ERR_NOACK);
-
-	if (memcmp(acme_alpn_name, in, sizeof(acme_alpn_name)))
-		return (SSL_TLSEXT_ERR_NOACK);
-
-	*out = in + 1;
-	*outlen = inlen - 1;
-
-	c->flags |= CONN_TLS_ALPN_ACME_SEEN;
-
-	/*
-	 * If SNI was already done, we can continue, otherwise we mark
-	 * that we saw the right ALPN negotiation on this connection
-	 * and wait for the SNI extension to be parsed.
-	 */
-	if (c->flags & CONN_TLS_SNI_SEEN) {
-		/* SNI was seen, we are on the right domain. */
-		kore_acme_tls_challenge_use_cert(ssl, udata);
-	}
-
-	return (SSL_TLSEXT_ERR_OK);
-}
-
 void
 kore_acme_get_paths(const char *domain, char **key, char **cert)
 {
@@ -416,43 +376,6 @@ kore_acme_get_paths(const char *domain, char **key, char **cert)
 	*key = kore_strdup(path);
 }
 
-void
-kore_acme_tls_challenge_use_cert(SSL *ssl, struct kore_domain *dom)
-{
-	struct connection	*c;
-	const unsigned char	*ptr;
-	X509			*x509;
-
-	if (dom->acme == 0) {
-		kore_log(LOG_NOTICE, "[%s] ACME not active", dom->domain);
-		return;
-	}
-
-	if (dom->acme_challenge == 0) {
-		kore_log(LOG_NOTICE,
-		    "[%s] ACME auth challenge not active", dom->domain);
-		return;
-	}
-
-	kore_log(LOG_NOTICE, "[%s] acme-tls/1 challenge requested",
-	    dom->domain);
-
-	if ((c = SSL_get_ex_data(ssl, 0)) == NULL)
-		fatal("%s: no connection data present", __func__);
-
-	ptr = dom->acme_cert;
-	if ((x509 = d2i_X509(NULL, &ptr, dom->acme_cert_len)) == NULL)
-		fatal("d2i_X509: %s", ssl_errno_s);
-
-	if (SSL_use_certificate(ssl, x509) == 0)
-		fatal("SSL_use_certificate: %s", ssl_errno_s);
-
-	SSL_clear_chain_certs(ssl);
-	SSL_set_verify(ssl, SSL_VERIFY_NONE, NULL);
-
-	c->proto = CONN_PROTO_ACME_ALPN;
-}
-
 static void
 acme_parse_directory(void)
 {
@@ -483,7 +406,7 @@ acme_parse_directory(void)
 	if (!kore_json_parse(&json)) {
 		kore_log(LOG_NOTICE,
 		    "failed to parse directory payload from ACME server (%s)",
-		    kore_json_strerror(&json));
+		    kore_json_strerror());
 		goto cleanup;
 	}
 
@@ -688,7 +611,7 @@ acme_order_create_submit(struct acme_sign_op *op, struct kore_buf *payload)
 	if (!kore_json_parse(&json)) {
 		kore_log(LOG_NOTICE,
 		    "[%s] failed to parse order payload from ACME server (%s)",
-		    domain, kore_json_strerror(&json));
+		    domain, kore_json_strerror());
 		goto cleanup;
 	}
 
@@ -806,7 +729,7 @@ acme_order_update_submit(struct acme_sign_op *op, struct kore_buf *payload)
 	if (!kore_json_parse(&json)) {
 		kore_log(LOG_NOTICE,
 		    "[%s] failed to parse order payload from ACME server (%s)",
-		    order->domain, kore_json_strerror(&json));
+		    order->domain, kore_json_strerror());
 		goto cleanup;
 	}
 
@@ -989,7 +912,7 @@ acme_order_remove(struct acme_order *order, const char *reason)
 		kore_free(auth);
 	}
 
-	kore_log(LOG_NOTICE, "[%s] order removed (%s)", order->domain, reason);
+	kore_log(LOG_INFO, "[%s] order removed (%s)", order->domain, reason);
 
 	if (strcmp(reason, "completed"))
 		acme_order_retry(order->domain);
@@ -1224,7 +1147,7 @@ acme_order_auth_update_submit(struct acme_sign_op *op, struct kore_buf *payload)
 	if (!kore_json_parse(&json)) {
 		kore_log(LOG_NOTICE,
 		    "[%s:auth] failed to parse payload from ACME server (%s)",
-		    order->domain, kore_json_strerror(&json));
+		    order->domain, kore_json_strerror());
 		goto cleanup;
 	}
 
@@ -1366,7 +1289,7 @@ cleanup:
 	} else {
 		order->auths--;
 		if (order->auths == 0) {
-			kore_log(LOG_NOTICE,
+			kore_log(LOG_INFO,
 			    "[%s:auth] authentications done", order->domain);
 			order->state = ACME_ORDER_STATE_RUNNING;
 		}
@@ -1389,12 +1312,12 @@ acme_challenge_tls_alpn_01(struct acme_order *order,
 		acme_challenge_tls_alpn_01_create(order, challenge);
 		break;
 	case ACME_STATUS_PROCESSING:
-		kore_log(LOG_NOTICE,
+		kore_log(LOG_INFO,
 		    "[%s:auth:challenge:tls-alpn-01] processing",
 		    order->domain);
 		break;
 	case ACME_STATUS_VALID:
-		kore_log(LOG_NOTICE,
+		kore_log(LOG_INFO,
 		    "[%s:auth:challenge:tls-alpn-01] valid",
 		    order->domain);
 		ret = KORE_RESULT_OK;
@@ -1419,7 +1342,7 @@ acme_challenge_tls_alpn_01_create(struct acme_order *order,
 	u_int8_t		digest[SHA256_DIGEST_LENGTH];
 
 	if (challenge->flags & ACME_FLAG_CHALLENGE_CREATED) {
-		kore_log(LOG_NOTICE,
+		kore_log(LOG_INFO,
 		    "[%s:auth:challenge:tls-alpn-01] pending keymgr",
 		    order->domain);
 		return;
@@ -1427,7 +1350,7 @@ acme_challenge_tls_alpn_01_create(struct acme_order *order,
 
 	challenge->flags |= ACME_FLAG_CHALLENGE_CREATED;
 
-	kore_log(LOG_NOTICE,
+	kore_log(LOG_INFO,
 	    "[%s:auth:challenge:tls-alpn-01] requested from keymgr",
 	    order->domain);
 
@@ -1613,7 +1536,7 @@ acme_sign_submit(struct kore_json_item *json, const char *url, void *udata,
 	op->id = signop_id++;
 	op->payload = acme_base64url(buf.data, buf.offset);
 	op->protected = acme_protected_component(op->nonce, url);
-	op->t = kore_timer_add(acme_sign_expire, 5000, op, KORE_TIMER_ONESHOT);
+	op->t = kore_timer_add(acme_sign_expire, 30000, op, KORE_TIMER_ONESHOT);
 
 	kore_buf_reset(&buf);
 	kore_buf_append(&buf, &op->id, sizeof(op->id));
@@ -1753,8 +1676,10 @@ acme_base64url(const void *data, size_t len)
 {
 	char		*b64;
 
-	if (!kore_base64url_encode(data, len, &b64, KORE_BASE64_RAW))
-		fatal("%s: failed to encode base64url data of %zu bytes", len);
+	if (!kore_base64url_encode(data, len, &b64, KORE_BASE64_RAW)) {
+		fatal("%s: failed to encode base64url data of %zu bytes",
+		    __func__, len);
+	}
 
 	return (b64);
 }
